/*
// author : Sven Goethel - Stapenhorststrasse 35a - 33615 Bielefeld
*/
#include <stdio.h>
#include <stdlib.h>

#include "rtti.h"
#include "cliste.h"
#include "strclass.h"
#include "c166mem.h"

typedef struct Int {
  rtti rtype;
  int a;
} Int;

static rtti *IntRtti=NULL;

typedef struct Real {
  rtti rtype;
  double a;
} Real;

static rtti *RealRtti=NULL;

typedef struct container
{
  rtti rtype;
  HCListe LInt;
  HCListe LStr;
} container;
static rtti *ContainerRtti=NULL;

void ResetContainer(void)
{
  ContainerRtti=NULL;
}

int    CmpContainer     ( const container *a , const container *b)
{
  return CmpRtti((rtti*)(a->LInt), (rtti*)(b->LInt)) &&
         CmpRtti((rtti*)(a->LStr), (rtti*)(b->LStr)) ;
}
rtti * CreateContainer  ( void )
{
  container * res= calloc(1, sizeof(container));
  if(!res) INT_ERR(__LINE__);
  if(!ContainerRtti) INT_ERR(__LINE__);
  res->LInt=CreateRtti(GetRttiCListe());
  res->LStr=CreateRtti(GetRttiCListe());
  res->rtype.type=ContainerRtti->type;
  res->rtype.sz=ContainerRtti->sz;
  return C_RTTI(res);
}
rtti * CopyCreateContainer ( const container * a)
{
  container * res= calloc(1, sizeof(container));
  if(!res) INT_ERR(__LINE__);
  if(!a) INT_ERR(__LINE__);
  if(!ContainerRtti) INT_ERR(__LINE__);
  res->LInt=CopyCreateRtti((rtti *)(a->LInt));
  res->LStr=CopyCreateRtti((rtti *)(a->LStr));
  res->rtype.type=ContainerRtti->type;
  res->rtype.sz=ContainerRtti->sz;
  return C_RTTI(res);
}

void   DestroyContainer    ( container * a)
{
  if(!a) INT_ERR(__LINE__);
  if(!ContainerRtti) INT_ERR(__LINE__);
  DestroyRtti((rtti *)(a->LInt));
  DestroyRtti((rtti *)(a->LStr));
  free(a);
}

void ContainerRegisterRtti       (void)
{
  if(ContainerRtti == NULL)
  {
    ContainerRtti = NewRttiType(sizeof(container));
    RegisterRttiFct(ContainerRtti,
                    (T_FCmpRtti) CmpContainer,
                    (T_FCreateRtti) CreateContainer,
                    (T_FCopyCreateRtti) CopyCreateContainer,
                    (T_FDestroyRtti) DestroyContainer,
										(T_FResetRtti) ResetContainer
                   );
  }
}

void _HeapCheck(int line)
{
   int heap=0;
   if((heap=HeapCheck()) != HEAP_OK)
   { ErrorOut("HeapError %d\n\r", heap);INT_ERR(line); }
}

#ifndef _C166_
	#define OUTPUT printf
#else
	#define OUTPUT ErrorOut
#endif

int TestListe(void)
{
  HCListe h1,h2,h3,h4, h5, h6, h7, h8;
  container *pc1=NULL;
  Int *pn1=NULL;
  Real *pm1=NULL;
  int i, j, k, heap;
  StrClass *ps1=NULL;

  OUTPUT("Start CListe-Test (incl. RTTI)\n\r");

  _HeapCheck(__LINE__);

  (void) StrClassRegisterRtti();
  ps1=(StrClass *)CreateRtti(GetRttiStrClass());

  _HeapCheck(__LINE__);

  CListeRegisterRtti       ();
  IntRtti=NewRttiType(sizeof(Int));
  RealRtti=NewRttiType(sizeof(Real));

  _HeapCheck(__LINE__);

  pn1=(Int *)CreateRtti(IntRtti);
  pm1=(Real *)CreateRtti(RealRtti);

  _HeapCheck(__LINE__);

  ContainerRegisterRtti       ();
  pc1=(container *)CreateRtti(ContainerRtti);

  _HeapCheck(__LINE__);

  h1=CreateRtti(GetRttiCListe());
  h2=CreateRtti(GetRttiCListe());
  h3=CreateRtti(GetRttiCListe());
  h4=NULL;
  h5=CreateRtti(GetRttiCListe());
  h6=NULL;
  h7=CreateRtti(GetRttiCListe());
  h8=NULL;

  _HeapCheck(__LINE__);

  for(i=0; i<10; i++)
  {
    StrClassPrintf        (ps1, "StrClass element %d\n\r", i+1);
    CListe_FuegeEin(h5, (rtti *)ps1, i);
    if( !CListe_IstElement(h5, (rtti *)ps1, 0)<0 )
      OUTPUT("s1 !e h5 => nanu ?\n\r");
  }
  _HeapCheck(__LINE__);

  pc1->LStr=(HCListe *)AssignRtti((rtti *)(pc1->LStr), (rtti *)h5);

  _HeapCheck(__LINE__);

  for(i=0; i<10; i++)
  {
    pn1->a=i;
    CListe_FuegeEin(h1, (rtti *)pn1, 0);
    if( !CListe_IstElement(h1, (rtti *)pn1, 0)<0 )
      OUTPUT("n1 !e h1 => nanu ?\n\r");
  }
  _HeapCheck(__LINE__);

  pc1->LInt=(HCListe *)AssignRtti((rtti *)(pc1->LInt), (rtti *)h1);

  _HeapCheck(__LINE__);

  CListe_FuegeEin(h7, (rtti *)pc1, 0);
  CListe_FuegeEin(h7, (rtti *)pc1, 1);
  _HeapCheck(__LINE__);

  h8=CopyCreateRtti((rtti *)h7);
  _HeapCheck(__LINE__);

  if( !CmpRtti((rtti *)h7, (rtti *)h8) )
    OUTPUT("h7 != h8 => nanu ?\n\r");
  _HeapCheck(__LINE__);

  CListe_Loesche(h8, 0);
  _HeapCheck(__LINE__);

  if( CmpRtti((rtti *)h7, (rtti *)h8) )
    OUTPUT("h7 == h8 => nanu ?\n\r");
  _HeapCheck(__LINE__);

  DestroyRtti((rtti *)pc1);
  DestroyRtti((rtti *)h7);
  DestroyRtti((rtti *)h8);

  _HeapCheck(__LINE__);

  i=0;
  while(i<10)
  {
    pn1->a=i;
    CListe_FuegeEin(h2, (rtti *)pn1, i);
    if( CListe_IstElement(h2, (rtti *)pn1, 0)<0 )
      OUTPUT("n1 !e h2 => nanu ?\n\r");
    i++;
    pm1->a=i+(double)i/10;
    CListe_FuegeEin(h2, (rtti *)pm1, i);
    if( !CListe_IstElement(h2, (rtti *)pm1, 0)<0 )
      OUTPUT("m1 !e h2 => nanu ?\n\r");
    i++;
  }

  _HeapCheck(__LINE__);

  h3=(HCListe)AssignRtti((rtti *)h3,(rtti *)h1);
  h4=CopyCreateRtti((rtti *)h2);
  h6=CopyCreateRtti((rtti *)h5);

  _HeapCheck(__LINE__);


  if( !CmpRtti((rtti *)h5, (rtti *)h6) )
    OUTPUT("h5 != h6 => nanu ?\n\r");

  CPutStrClass((StrClass *)CListe_GetElement(h6, 0), "modified element 0\n\r");

  _HeapCheck(__LINE__);

  for(i=0; i<10; i++)
  {
    if( CListe_GetElement(h6, i)->type != GetRttiStrClass()->type )
      INT_ERR(__LINE__);
    printf("assigned liste h6 element %d : %s\n\r",
      i, ((StrClass *)(CListe_GetElement(h6, i)))->StrPtr);
  }
  for(i=0; i<10; i++)
  {
    if( CListe_GetElement(h3, i)->type != IntRtti->type ) INT_ERR(__LINE__);
    printf("assigned liste h3 element %d : %d\n\r",
      i, ((Int *)(CListe_GetElement(h3, i)))->a);
  }
  for(i=0; i<10; i++)
  {
    printf("Liste h4[%d] type = ", i);
    if( CListe_GetElement(h4, i)->type == RealRtti->type )
      printf("Real\n\r");
    else
      printf("Int\n\r");
  }


  _HeapCheck(__LINE__);

  i=0;
  while(i<10)
  {
    if( CListe_GetElement(h4, i)->type != IntRtti->type ) INT_ERR(__LINE__);
    printf("copycreate liste h4 element %d : %d\n\r",
      i, ((Int *)(CListe_GetElement(h4, i)))->a);
    i++;
    if( CListe_GetElement(h4, i)->type != RealRtti->type ) INT_ERR(__LINE__);
    printf("copycreate liste h4 element %d : %lf\n\r",
      i, ((Real *)(CListe_GetElement(h4, i)))->a);
    i++;
  }

  _HeapCheck(__LINE__);

  if( CmpRtti((rtti *)h5, (rtti *)h6) )
    OUTPUT("h5 == h6 => nanu ?\n\r");

  if( !CmpRtti((rtti *)h1, (rtti *)h3) )
    OUTPUT("h1 != h3 => nanu ?\n\r");

  if( !CmpRtti((rtti *)h2, (rtti *)h4) )
    OUTPUT("h2 != h4 => nanu ?\n\r");

  _HeapCheck(__LINE__);

  DestroyRtti((rtti *)h1);
  DestroyRtti((rtti *)h2);
  DestroyRtti((rtti *)h3);
  DestroyRtti((rtti *)h4);
  DestroyRtti((rtti *)h5);
  DestroyRtti((rtti *)h6);

  _HeapCheck(__LINE__);

  DestroyRtti((rtti *)pm1);
  DestroyRtti((rtti *)pn1);
  DestroyRtti((rtti *)ps1);

  _HeapCheck(__LINE__);

  ResetRtti();

  _HeapCheck(__LINE__);

  OUTPUT("If no error-message occured\n\rCListen-Test was succesfull !");

  return 0;
}

