
#include "mathp.h"

double atan2_coord(t_coord sp, t_coord ep){
    if (sp.x==ep.x && sp.y==ep.y) return(0.0);
    return(atan2(sp.y.value()-ep.y.value(), sp.x.value()-ep.x.value()));
}

t_real distance(const t_coord &sp, const t_coord &ep){
    t_real dx;
    dx = ep.x - sp.x;
    return sqrt(sqr((ep.x-sp.x))+sqr(ep.y-sp.y)+sqr(ep.z-sp.z));
}

/*DOCST-------------------------------------------------------------------------

    Definition und Durchfuehrung von Koordinatentransformationen
    =================

    - Beschreibung
    --------------
    ???
           ( M00 M01 M02 M03 )
    Matrix M = ( M10 M11 M12 M13 )
           ( M20 M21 M22 M23 )
           ( M30 M31 M32 M33 )

    - Eingabe-Parameter
    -----------
    E:  ????                ???


    - Ausgabe-Parameter
    -----------
    A:  ????                ???

    - Returncode
    ------------
    ????

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    stderr          wird fuer Fehlermeldungen benutzt

-------------------------------------------------------------------------DOCEND*/
void transform_coord( const t_orient &or, t_coord & point ) 
{
  t_transformation_matrix matrix;
	orientation_to_matrix(or, matrix);
  transform_coord( matrix, point );
}

void orientation_to_matrix ( const t_orient &or, t_transformation_matrix &goal )
{
    long double sa, sb, sc;
    long double ca, cb, cc;

    sa = sin(or.a.value());
    sb = sin(or.b.value());
    sc = sin(or.c.value());
    ca = cos(or.a.value());
    cb = cos(or.b.value());
    cc = cos(or.c.value());

    goal[0][0] = cc*cb;
    goal[0][1] = -ca*sc+sa*sb*cc;
    goal[0][2] =  sa*sc+ca*sb*cc;
    goal[0][3] = 0.0;
    goal[1][0] = cb*sc;
    goal[1][1] = ca*cc+sa*sb*sc;
    goal[1][2] = -sa*cc+ca*sb*sc;
    goal[1][3] = 0.0;
    goal[2][0] = -sb;
    goal[2][1] = sa*cb;
    goal[2][2] = ca*cb;
    goal[2][3] = 0.0;
    goal[3][0] = 0.0;
    goal[3][1] = 0.0;
    goal[3][2] = 0.0;
    goal[3][3] = 1.0;
}

void transform_coord( const t_transformation_matrix &matrix, t_coord & point ) 
{
    double src[4], trg[4];
    int i, j;

    src[0] = point.x.value();
    src[1] = point.y.value();
    src[2] = point.z.value();
    src[3] = 1.0;

    for (i=0; i<4; i++) trg[i] = 0.0;

    for (i=0; i<4; i++)
        for (j=0; j<4; j++)
            trg[i] += matrix[i][j] * src[j];

    point.x = trg[0];
    point.y = trg[1];
    point.z = trg[2];
}

/* Die Funktion transform_orientation nimmt ... */
int matrix_to_orientation(
        const t_transformation_matrix &matrix,
        t_orient &or_1, t_orient &or_2,
        t_coord &sin_1, t_coord &sin_2,
        t_coord &cos_1, t_coord &cos_2) 
{
    t_orient a[2];
    t_coord s[2], c[2];
    int i, noswap;

    s[0].y = - matrix[2][0];
    s[1].y = s[0].y;
    c[0].y = sqrt(1-s[0].y*s[0].y);
    c[1].y = c[0].y*(-1);
    a[0].b = c[0].y<0 ? M_PI-asin(s[0].y).value() : asin(s[0].y).value();
    a[1].b = c[1].y<0 ? M_PI-asin(s[1].y).value() : asin(s[1].y).value();
    if (c[0].y==0.0) {
        s[0].x = s[1].x = 0.0;
        c[0].x = c[1].x = 1.0;
        a[0].a = a[1].a = 0.0;
        s[0].z = s[1].z = - matrix[0][1];
        c[0].z = c[1].z = matrix[0][2];
        a[0].c = a[1].c
            = c[1].z<0 ? M_PI-asin(s[1].z).value() : asin(s[1].z).value();
    } else { /* cb != 0.0 */
        for (i=0; i<2; i++) {
        s[i].x = matrix[2][1] / c[i].y;
        c[i].x = matrix[2][2] / c[i].y;
        a[i].a  = c[i].x<0 ? M_PI-asin(s[i].x).value() : asin(s[i].x).value();
        s[i].z = matrix[1][0] / c[i].y.value();
        c[i].z = matrix[0][0] / c[i].y.value();
        a[i].c = c[i].z<0 ? M_PI-asin(s[i].z).value() : asin(s[i].z).value();
        }
    }

    noswap = ((a[1].a+a[1].b+a[1].c)>=(a[0].a+a[0].b+a[0].c));
    or_1=a[1-noswap];
    or_2=a[1-noswap];
    sin_1=s[1-noswap];
    sin_2=s[1-noswap];
    cos_1=c[1-noswap];
    cos_2=c[1-noswap];

    return (or_1==or_2) ? 1 : 2;
}

void matrix_multiplication( const t_transformation_matrix &mat1,
														const t_transformation_matrix &mat2,
														t_transformation_matrix &goal)
{
    int i, j, k;

    for (i=0; i<4; i++)
        for (j=0; j<4; j++)
            goal[i][j] = 0.0;

    for (i=0; i<4; i++)
        for (j=0; j<4; j++)
        for (k=0; k<4; k++)
            goal[i][j] += mat1[i][k]*mat2[k][j];
}

void or_off_to_matrix( const t_orient &orient, const t_coord &origin, 
                       t_transformation_matrix& goal )
{
    orientation_to_matrix(orient, goal);
    goal[0][3] = origin.x.value();
    goal[1][3] = origin.y.value();
    goal[2][3] = origin.z.value();
}

/*DOCST-------------------------------------------------------------------------

    Filter zur Umsetzung der Orientierungen
    =================

    - Beschreibung
    --------------
    Das Programm nimmt die Orientierungen der FMX-Datenstruktur,
    die eine Drehung erst um die X-Achse, dann um die Y-Achse und
    zuletzt um die Z-Achse beschreiben, und wandelt diese um in
    Neigungs- und Drehwinkel.  Der Neigungswinkel beschreibt eine
    Neigung um die X-Achse, der Drehwinkel beschreibt eine Drehung
    um die Z-Achse.

    - Eingabe-Parameter
    -----------
    E:  argc                Anzahl der von BS uebergebenen Parameter
    E:  argv                Zeiger auf ein Feld von Strings der Parameter

    - Ausgabe-Parameter
    -----------
    A:  ----                keine

    - Returncode
    ------------
    immer 0

    - allgemeine Ein- und Ausgabeeinheiten
    --------------------------------------
    stdin           wird nicht benutzt
    stdout          wird nicht benutzt
    stderr          wird fuer Fehlermeldungen benutzt

-------------------------------------------------------------------------DOCEND*/

double BOGEN(t_angle x){
    return x.value()*180/M_PI;
}

t_coord trafo_xyz2incl(t_orient &or)
{
    t_transformation_matrix matrix;
    t_coord sample ( 0.0, 0.0, -1.0 );
    t_angle rot = 0.0;
    t_angle incl = 0.0;
    double length = 0.0;

    orientation_to_matrix(or, matrix);
    transform_coord(matrix, sample);

    length = sqrt((sample.x*sample.x+sample.y*sample.y).value());
    incl = asin(length);
    if (sample.z<0.0) incl = M_PI-incl.value();

    if (sample.x==0.0 && sample.y==0.0)
        rot = 0.0;
    else
        rot = atan2((sample.y.value()), (sample.x.value())) + M_PI_2;
    if (rot >   M_PI) rot -= 2*M_PI;
    if (rot <= -M_PI) rot += 2*M_PI;
    if (rot>M_PI_2 || rot<-M_PI_2){
        incl = -incl.value();
        rot = rot + M_PI;
    }
    if (rot >   M_PI) rot -= 2*M_PI;

#ifdef DEBUG
    fprintf(stderr, "or=(%.4f,%.4f,%.4f)",
        BOGEN(or->a), BOGEN(or->b), BOGEN(or->c));
    fprintf(stderr, ", # dx=%6.3f dy=%6.3f dz=%6.3f",
        sample.x.value(), sample.y.value(), sample.z.value());
    fprintf(stderr, ", # Neigung=%.4f, Drehung=%.4f\n",
        BOGEN(incl), BOGEN(rot));
    fprintf(stderr,"<CR>\n");
    getchar();
#endif
    or.a = incl;
    or.b = rot;
    or.c = 0.0;
    return sample;
}

